from django.test import TestCase

from r_shepard.api.models import Container, Project


class ProjectModelTest(TestCase):
    def setUp(self):
        self.project = Project.objects.create(
            name="Test Project",
            slug="test-project",
            data_path="/path/to/data",
            workspace_path="/path/to/workspace",
            max_containers=4,
            max_ram=32,
            max_cpus=8,
            auto_commit_enabled=False,
            git_repo_url="https://github.com/user/repo.git",
            commit_interval=15,
        )

    def test_project_creation(self):
        self.assertIsInstance(self.project, Project)
        self.assertEqual(self.project.__str__(), "Test Project")

    def test_generate_slug(self):
        self.assertEqual(self.project.generate_slug(), "test_project")


class ContainerModelTest(TestCase):
    def setUp(self):
        self.project = Project.objects.create(
            name="Test Project",
            slug="test-project",
            data_path="/path/to/data",
            workspace_path="/path/to/workspace",
            max_containers=4,
            max_ram=32,
            max_cpus=8,
            auto_commit_enabled=False,
            git_repo_url="https://github.com/user/repo.git",
            commit_interval=15,
        )
        self.container = Container.objects.create(
            project=self.project,
            image="rocker/rstudio",
            tag="latest",
            password="password",
            ram=2,
            cpus=2,
            port=8787,
            local_url="http://localhost:8787",
            tailscale_funnel_url="http://funnel.url",
        )

    def test_container_creation(self):
        self.assertIsInstance(self.container, Container)
        self.assertEqual(self.container.__str__(), self.container.container_id)

    def test_generate_container_id(self):
        container_id = Container.generate_container_id()
        self.assertEqual(len(container_id), 8)



# Generated by Django 5.0.4 on 2024-04-24 13:21

import r_shepard.api.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0014_project_max_containers"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="auto_commit_enabled",
            field=models.BooleanField(
                default=False,
                help_text="Whether automatic commits are enabled for the workspace_path.",
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="git_repo_url",
            field=models.URLField(
                blank=True,
                help_text="URL of the Git repository.",
                max_length=255,
                null=True,
                validators=[r_shepard.api.models.validate_git_url],
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-22 07:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0008_container_port_alter_container_container_id_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='container',
            name='port',
            field=models.PositiveIntegerField(default=8787, help_text="The port on which the container's RStudio instance is accessible."),
        ),
    ]



import re
import subprocess

from podman import PodmanClient

from .models import Container
from .podman import HOST_IP, PODMAN_SOCKET


def tailscale_serve(port: int, bg: bool = True, path: str = None):
    command = "tailscale serve"
    if bg:
        command += " --bg"
    if path:
        command += f" --set-path={path}"
    command += f" {port}"
    result = subprocess.run(command, shell=True, capture_output=True, text=True)

    # If result.stderr contains "Tailscale is stopped", then start it
    if result.stderr and "Tailscale is stopped" in result.stderr:
        subprocess.run("tailscale up", shell=True)
        result = subprocess.run(command, shell=True, capture_output=True, text=True)

    # Extract URL from output
    match = re.search(r"(https://\S+)", result.stdout)
    if match:
        url = match.group(1)
        print(f"Container served within tailnet at {url}")
        return url


def serve_container(container: Container):
    path = f"/{container.project.slug}/{container.container_id}"

    # Apply path to RStudio configuration
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        podman_container = client.containers.get(
            f"rstudio_{container.project.slug}_{container.container_id}"
        )
        podman_container.exec_run(
            cmd=[
                "/bin/sh",
                "-c",
                f"if ! grep -q 'www-root-path={path}' /etc/rstudio/rserver.conf; then echo 'www-root-path={path}' >> /etc/rstudio/rserver.conf; fi",
            ],
            privileged=True,
            detach=True,
        )
        podman_container.restart()

    container.tailscale_serve_url = tailscale_serve(container.port, True, path=path)
    container.local_url = ""
    container.save()


def stop_serve_container(container: Container):
    path = f"/{container.project.slug}/{container.container_id}"
    escaped_path = path.replace("/", "\\/")
    print(escaped_path)
    command = f"tailscale serve {path} off"
    subprocess.run(command, shell=True)
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        podman_container = client.containers.get(
            f"rstudio_{container.project.slug}_{container.container_id}"
        )
        podman_container.exec_run(
            cmd=[
                "/bin/sh",
                "-c",
                f"if grep -q 'www-root-path={escaped_path}' /etc/rstudio/rserver.conf; then sed -i '/www-root-path={escaped_path}/d' /etc/rstudio/rserver.conf; fi",
            ],
            privileged=True,
            detach=True,
        )
        podman_container.restart()

    container.tailscale_serve_url = ""
    container.local_url = f"http://{HOST_IP}:{container.port}"
    container.save()


def tailscale_funnel(port: int, bg: bool = True, path: str = None):
    command = "tailscale funnel"
    if bg:
        command += " --bg"
    if path:
        command += f" --set-path={path}"
    command += f" {port}"
    subprocess.run(command, shell=True)



# Generated by Django 5.0.4 on 2024-04-20 16:40

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of the project.', max_length=255, unique=True)),
                ('data_path', models.CharField(help_text='Filesystem path where project data is stored.', max_length=1024)),
                ('workspace_path', models.CharField(help_text='Filesystem path where R scripts are stored.', max_length=1024)),
                ('password', models.CharField(help_text='Password for the RStudio Rocker container.', max_length=255)),
                ('ram', models.PositiveIntegerField(help_text='Amount of RAM in MB allocated for the project.')),
                ('cpus', models.PositiveIntegerField(help_text='Number of CPUs allocated to the project.')),
            ],
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-22 08:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0010_alter_container_port'),
    ]

    operations = [
        migrations.AlterField(
            model_name='container',
            name='tailscale_serve_url',
            field=models.URLField(default='', help_text="Tailscale Serve URL to access this container's RStudio instance from the tailnet", max_length=1024, null=True),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-24 13:32

import datetime
import r_shepard.api.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0015_project_auto_commit_enabled_project_git_repo_url"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="commit_interval",
            field=models.PositiveIntegerField(
                default=60, help_text="Commit interval in seconds."
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="last_commit_time",
            field=models.DateTimeField(
                auto_now_add=True,
                default=datetime.datetime(
                    2024, 4, 24, 13, 32, 6, 347026, tzinfo=datetime.timezone.utc
                ),
                help_text="Last time the project was committed to the Git repository.",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="project",
            name="git_repo_url",
            field=models.CharField(
                blank=True,
                help_text="URL of the Git repository.",
                max_length=255,
                null=True,
                validators=[r_shepard.api.models.validate_git_url],
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-20 17:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='is_active',
            field=models.BooleanField(default=False, help_text="Indicates whether the project's container is running"),
        ),
    ]



import random
import re
import string

from django.contrib.auth.hashers import check_password, make_password
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone

# Create your models here.


def validate_git_url(value):
    pattern = r"(git@|http:\/\/|https:\/\/)([\w\.@:]+)(\/|:)([\w\.\/\-]+\.git)"
    if not re.match(pattern, value):
        raise ValidationError(f"{value} is not a valid Git URL")


class Project(models.Model):
    name = models.CharField(
        max_length=255,
        unique=True,
        help_text="The name of the project.",
    )
    slug = models.SlugField(
        max_length=255,
        unique=True,
        help_text="A URL-safe slug for the project.",
    )
    data_path = models.CharField(
        max_length=1024,
        help_text="Filesystem path where project data is stored. Containers have read access here.",
    )
    workspace_path = models.CharField(
        max_length=1024,
        help_text="Filesystem path where R scripts are stored. The containers have write access here.",
    )
    max_containers = models.PositiveIntegerField(
        default=4,
        help_text="Maximum number of containers that can be run for the project.",
    )
    max_ram = models.PositiveIntegerField(
        default=32,
        help_text="Maximum amount of RAM in GB allocated for the project.",
    )
    max_cpus = models.PositiveIntegerField(
        default=8,
        help_text="Maximum number of CPUs allocated to the project.",
    )
    auto_commit_enabled = models.BooleanField(
        default=False,
        help_text="Whether automatic commits are enabled for the workspace_path.",
    )
    git_repo_url = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        validators=[validate_git_url],
        help_text="URL of the Git repository.",
    )
    commit_interval = models.PositiveIntegerField(
        default=15, help_text="Commit interval in minutes."
    )
    last_commit_time = models.DateTimeField(
        default=timezone.now,
        help_text="Last time the project was committed to the Git repository.",
    )

    # Generate a URL-safe slug for the project
    def generate_slug(self):
        """Generate a URL-safe slug from the project name."""
        return self.name.lower().replace(" ", "_")

    def running_containers(self):
        return self.containers.filter(is_running=True)

    def __str__(self):
        return self.name


class Container(models.Model):
    class Meta:
        unique_together = ("project", "container_id")

    def generate_container_id():
        """Generate a random 8-char string."""
        return "".join(random.choices(string.ascii_lowercase + string.digits, k=8))

    project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="containers",
        help_text="The project to which this container belongs.",
    )
    container_id = models.CharField(
        max_length=8,
        default=generate_container_id,
        verbose_name="Container ID",
        help_text="The unique identifier of the container within the project.",
    )
    image = models.CharField(
        default="rocker/rstudio",
        max_length=255,
        help_text="The name of the container image.",
    )
    tag = models.CharField(
        default="latest",
        max_length=255,
        help_text="The tag of the container image.",
    )
    password = models.CharField(
        max_length=255,
        help_text="Password for the RStudio Rocker container.",
    )
    ram = models.PositiveIntegerField(
        default=2,
        verbose_name="RAM Allocation (GB)",
        help_text="The maximum amount of RAM in GB the container can use. Will only take effect after restarting the container.",
    )
    cpus = models.PositiveIntegerField(
        default=2,
        verbose_name="CPU Allocation",
        help_text="The maximum number of CPU cores the container can use. Will only take effect after restarting the container.",
    )
    port = models.PositiveIntegerField(
        default=8787,
        verbose_name="Host Port",
        help_text="The host port on which the container's RStudio instance is accessible locally.",
    )
    local_url = models.URLField(
        default="",
        max_length=1024,
        verbose_name="Local URL",
        help_text="Local URL to access this container's RStudio instance",
    )
    tailscale_funnel_url = models.URLField(
        default="",
        max_length=1024,
        verbose_name="Tailscale Funnel URL",
        help_text="Tailscale Funnel URL to access this container's RStudio instance from the web",
    )
    tailscale_serve_url = models.URLField(
        default="",
        null=True,
        max_length=1024,
        verbose_name="Tailscale Serve URL",
        help_text="Tailscale Serve URL to access this container's RStudio instance from the tailnet",
    )
    is_running = models.BooleanField(
        default=False,
        verbose_name="Running",
        help_text="Indicates whether the container is running",
    )

    def save(self, *args, **kwargs):
        if not check_password(self.password, self.__original_password):
            self.password = make_password(self.password)
        super().save(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__original_password = self.password

    def __str__(self):
        return self.container_id



# Generated by Django 5.0.4 on 2024-04-23 09:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0011_alter_container_tailscale_serve_url"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="slug",
            field=models.SlugField(
                default="test",
                help_text="A URL-safe slug for the project.",
                max_length=255,
                unique=True,
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-22 05:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0006_remove_project_cpus_remove_project_is_active_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='container',
            name='container_id',
            field=models.CharField(help_text='The unique identifier of the container.', max_length=5, unique=True),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-20 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0003_project_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='rstudio_version',
            field=models.CharField(default='latest', help_text='The version of the RStudio Rocker container.', max_length=255),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-21 18:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0005_remove_project_url_project_local_url_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='project',
            name='cpus',
        ),
        migrations.RemoveField(
            model_name='project',
            name='is_active',
        ),
        migrations.RemoveField(
            model_name='project',
            name='local_url',
        ),
        migrations.RemoveField(
            model_name='project',
            name='ram',
        ),
        migrations.RemoveField(
            model_name='project',
            name='rstudio_version',
        ),
        migrations.RemoveField(
            model_name='project',
            name='tailscale_funnel_url',
        ),
        migrations.RemoveField(
            model_name='project',
            name='tailscale_serve_url',
        ),
        migrations.AddField(
            model_name='project',
            name='max_cpus',
            field=models.PositiveIntegerField(default=8, help_text='Maximum number of CPUs allocated to the project.'),
        ),
        migrations.AddField(
            model_name='project',
            name='max_ram',
            field=models.PositiveIntegerField(default=32, help_text='Maximum amount of RAM in GB allocated for the project.'),
        ),
        migrations.CreateModel(
            name='Container',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.CharField(default='rocker/rstudio', help_text='The name of the container image.', max_length=255)),
                ('tag', models.CharField(default='latest', help_text='The tag of the container image.', max_length=255)),
                ('container_id', models.CharField(default='4Dj19', help_text='The unique identifier of the container.', max_length=5, unique=True)),
                ('ram', models.PositiveIntegerField(default=2, help_text='Amount of RAM in GB allocated for the container.')),
                ('cpus', models.PositiveIntegerField(default=2, help_text='Number of CPUs allocated to the container.')),
                ('local_url', models.URLField(default='', help_text="Local URL to access this container's RStudio instance", max_length=1024)),
                ('tailscale_funnel_url', models.URLField(default='', help_text="Tailscale Funnel URL to access this container's RStudio instance from the web", max_length=1024)),
                ('tailscale_serve_url', models.URLField(default='', help_text="Tailscale Serve URL to access this container's RStudio instance from the tailnet", max_length=1024)),
                ('is_running', models.BooleanField(default=False, help_text='Indicates whether the container is running')),
                ('project', models.ForeignKey(help_text='The project to which this container belongs.', on_delete=django.db.models.deletion.CASCADE, related_name='containers', to='api.project')),
            ],
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-21 17:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0004_project_rstudio_version'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='project',
            name='url',
        ),
        migrations.AddField(
            model_name='project',
            name='local_url',
            field=models.URLField(default='', help_text="Local URL to access the project's RStudio instance", max_length=1024),
        ),
        migrations.AddField(
            model_name='project',
            name='tailscale_funnel_url',
            field=models.URLField(default='', help_text="Tailscale Funnel URL to access the project's RStudio instance from the web", max_length=1024),
        ),
        migrations.AddField(
            model_name='project',
            name='tailscale_serve_url',
            field=models.URLField(default='', help_text="Tailscale Serve URL to access the project's RStudio instance from the tailnet", max_length=1024),
        ),
    ]



"""
Django settings for r_shepard project.

Generated by 'django-admin startproject' using Django 5.0.4.

For more information on this file, see
https://docs.djangoproject.com/en/5.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.0/ref/settings/
"""

import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-yc+#%@i2_hyrl1ut+c-9#ejwrv6(cv2de=-2+dv375a+oal*n7"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = os.getenv("ALLOWED_HOSTS", "").split(",")
CSRF_TRUSTED_ORIGINS = os.getenv("CSRF_TRUSTED_ORIGINS", "").split(",")

LOGIN_URL = "two_factor:login"

# this one is optional
LOGIN_REDIRECT_URL = "two_factor:profile"

# Application definition
INSTALLED_APPS = [
    "r_shepard.api",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_otp",
    "otp_yubikey",
    "django_otp.plugins.otp_static",
    "django_otp.plugins.otp_totp",
    # "django_otp.plugins.otp_email",  # <- if you want email capability.
    "two_factor",
    # "two_factor.plugins.phonenumber",  # <- if you want phone number capability.
    # "two_factor.plugins.email",  # <- if you want email capability.
    # "two_factor.plugins.yubikey",  # <- for yubikey capability.
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django_otp.middleware.OTPMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "r_shepard.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "r_shepard.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.0/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.environ.get("DB_PATH", BASE_DIR / "db.sqlite3"),
    }
}


# Password validation
# https://docs.djangoproject.com/en/5.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/5.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.0/howto/static-files/

STATIC_URL = "static/"
STATIC_ROOT = os.environ.get("STATIC_ROOT", "/home/r-shepard/static/")

# Default primary key field type
# https://docs.djangoproject.com/en/5.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Celery settings
CELERY_BROKER_URL = "redis://localhost:6379/0"
CELERY_ACCEPT_CONTENT = ["json"]
CELERY_TASK_SERIALIZER = "json"
CELERY_BEAT_SCHEDULE = {
    "auto_commit": {
        "task": "r_shepard.api.tasks.auto_commit",
        "schedule": 15,
    }
}



# Generated by Django 5.0.4 on 2024-04-23 11:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0012_project_slug"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="password",
        ),
        migrations.AddField(
            model_name="container",
            name="password",
            field=models.CharField(
                default="test",
                help_text="Password for the RStudio Rocker container.",
                max_length=255,
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-20 17:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0002_project_is_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='url',
            field=models.URLField(default='http://localhost:8387', help_text="URL to access the project's RStudio instance", max_length=1024),
            preserve_default=False,
        ),
    ]
